
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import ARDRegression
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', ARDRegression(
        tol = 0.001,
        n_iter = 300,
        fit_intercept = True,
        normalize = False,
        compute_score = False,
        copy_X = True,
        verbose = False,
    ))
])

hyper = {
    'lin__alpha_1': ss.expon(loc=1e-7, scale=.01),
    'lin__alpha_2': ss.expon(loc=1e-7, scale=.01),
    'lin__lambda_1': ss.expon(loc=1e-7, scale=.01),
    'lin__lambda_2': ss.expon(loc=1e-7, scale=.01),
    # 'lin__threshold_lambda': ss.randint(5, 10000),
}

meta = {
    'id': "simi14",
    'name': 'LinReg ARD',
    'descriptions': 'Automatic Relevance Determination',
    'solver': 'Evidence Maximization',
    'active': True,
    'keywords': [
        'linear regression', 'univariate regression', 'multiple regression',
        'bayesian regression', 'maximum likelihood estimation',
        'automatic relevance determination',
        'sparse bayesian learning and relevance vector machine',
        'sklearn.linear_model.ARDRegression'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}




from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import SGDRegressor
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', SGDRegressor(
        # Logistic Regression
        loss = 'squared_loss',
        penalty = 'l2',
        # l1_ratio = 0,
        fit_intercept = True,
        # solver settings
        max_iter = 1000,
        tol = 1e-3,
        shuffle = True,
        random_state = 42,
        # adaptive learning
        learning_rate = 'adaptive',
        eta0 = 0.5,
        # early stopping
        early_stopping = True,
        validation_fraction = 0.15,
        n_iter_no_change = 10,
        # other
        warm_start = True,
        average = False,  # averaging does not work!
    ))
])

hyper = {
    'lin__alpha': ss.gamma(a=1.2, loc=1e-6, scale=.08),  # alpha ~ [1e-6, 1]
}

meta = {
    'id': "simi11",
    'name': 'LinReg Ridge',
    'descriptions': (
        "Ridge Regression (L2 penalty), SGD solver, squared loss function."),
    'solver': 'Stochastic Gradient Descent',
    'active': True,
    'keywords': [
        'linear regression', 'univariate regression', 'multiple regression',
        'ridge', 'l2 penalty', 'sklearn.linear_model.SGDRegressor'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}




from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import HuberRegressor
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', HuberRegressor(
        fit_intercept = True,
        max_iter = 100,
        warm_start = True,
        tol = 1e-5
    ))
])

hyper = {
    'lin__epsilon': ss.uniform(1.0, 6.0),  # threshold for standardized error
    'lin__alpha': ss.gamma(a=1.5, loc=1e-5, scale=.7),  # alpha ~ [0.001, 10]
}

meta = {
    'id': "simi15",
    'name': 'LinReg Huber',
    'descriptions': (
        "Huber Regression, L-BFGS solver, standard-normal transformed "
        "features."),
    'solver': 'L-BFGS',
    'active': True,
    'keywords': [
        'robust regression', 'univariate regression', 'multiple regression',
        'huber regression', 'outlier', 'sklearn.linear_model.HuberRegressor',
        'huber', 'z-score'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}



# Store the best model somewhere
# https://scikit-learn.org/stable/modules/model_persistence.html
import os
import pickle
import joblib
import warnings


def to_disk(model, fname, path="tmp", tech="joblib"):
    # convert relative path to absolute path
    if not os.path.isabs(path):
        path = os.getcwd() + '/' + path
    # create folders
    os.makedirs(path, exist_ok=True)
    # add trailing slash
    path = os.path.join(path, '')
    # store it
    if tech is "joblib":
        joblib.dump(model, path + fname + '.joblib')
    elif tech is "pickle":
        pickle.dump(model, open(path + fname + '.pickle', 'wb'))
    else:
        raise Exception("unknown tech='...' specified.")
    return None


def from_disk(fname, path="tmp", tech="joblib"):
    # add trailing slash
    path = os.path.join(path, '')
    # load and return the model
    if tech is "joblib":
        filename = path + fname + '.joblib'
        if os.path.isfile(filename):  # check if file exist
            return joblib.load(filename)
    elif tech is "pickle":
        filename = path + fname + '.pickle'
        if os.path.isfile(filename):  # check if file exist
            return pickle.loads(open(filename, 'rb'))
    else:
        raise Exception("unknown tech='...' specified.")
    # file does not exist
    warnings.warn("'{:s}' does not not exist.".format(filename))
    return None



# For Single-Output Binary Classification

# Create Scorer - Ensure the highest value means "best"
import korr
from sklearn.metrics import (
    make_scorer,
    balanced_accuracy_score, hamming_loss, jaccard_similarity_score,
    precision_score, recall_score, matthews_corrcoef)


def scorerfun(y_true, y_pred):
    return korr.confusion_to_mcc(*korr.confusion(y_true, y_pred).ravel())


# Hyperparameter Optimization settings
cv_settings = {
    'n_iter': 50,  # How many random hyperparmeter settings to try
    'cv': 5,  # Number of CV folds
    'scoring': make_scorer(scorerfun),
    'random_state': 42,
    'n_jobs': -1,
    'return_train_score': True
}


# Evaluate by watching some metrics
def print_scores(y_true, y_pred):
    """
    Y_pred = model.predict(X_valid)
    print_scores(Y_valid, Y_pred)
    """
    print("{:>20s}: {:8.6f}".format(
        "Matthews", matthews_corrcoef(y_true, y_pred)))
    print("{:>20s}: {:8.6f}".format(
        "Accurancy", balanced_accuracy_score(y_true, y_pred)))
    print("{:>20s}: {:8.6f}".format(
        "Jaccard", jaccard_similarity_score(y_true, y_pred)))
    print("{:>20s}: {:8.6f}".format(
        "Hamming", hamming_loss(y_true, y_pred)))
    print("{:>20s}: {:8.6f}".format(
        "Precision", precision_score(y_true, y_pred)))
    print("{:>20s}: {:8.6f}".format(
        "Recall", recall_score(y_true, y_pred)))



from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import numpy as np


class PcaMinka(Pipeline):
    """Principal component analysis (PCA) with Minka's MLE
    to guess n_components.

    Example
    -------
        trans = PcaMinka()
        trans.fit(X_train)
        X_new = trans.transform(X_train)

    Notes
    -----
    The class 'PcaMinka' is a sklearn Pipeline and is equivalent to

        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.decomposition import PCA
        trans = Pipeline(steps=[
            ('scl', StandardScaler(with_mean=True, with_std=True, copy=True)),
            ('pca', PCA(
                n_components='mle', svd_solver='full',
                whiten=False, copy=True))
        ])
        trans.fit(X_train)
        X_new = trans.transform(X_train).astype(np.float16)
    """
    def __init__(self, prefix="minka"):
        self.prefix = prefix
        super().__init__(steps=[
            ('scl', StandardScaler(with_mean=True, with_std=True, copy=True)),
            ('pca', PCA(
                n_components='mle', svd_solver='full',
                whiten=False, copy=True))
        ])

    def transform(self, X, y=None):
        return super().transform(X).astype(np.float16)

    def fit(self, X, y=None):
        super().fit(X)
        self.feature_names_ = [
            self.prefix + "_" + str(i) for i in
            range(self.steps[1][1].n_components_)]
        return self


trans = PcaMinka()

meta = {
    'id': 'dim2',
    'name': 'PCA Minka MLE',
    'description': "Minka's MLE to guess n_components",
    'keywords': [
        'dimensionality reduction', 'principal component anlysis',
        'StandardScaler', 'PCA', 'Minka MLE']
}




from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import TheilSenRegressor
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', TheilSenRegressor(
        fit_intercept = True,
        copy_X = True,
        max_subpopulation = 10000,  # constrain 'n choose k' to 10k
        n_subsamples = None,  # maximal robustness
        max_iter = 300,
        tol = 1e-3,
        random_state = 42
    ))
])

hyper = {
    # 'lin__epsilon': ss.uniform(1.0, 6.0),  # threshold for standardized error
    # 'lin__alpha': ss.gamma(a=1.5, loc=1e-5, scale=.7),  # alpha ~ [0.001, 10]
}

meta = {
    'id': "simi16",
    'name': 'LinReg Theil-Sen',
    'descriptions': (
        "Theil-Sen Regression, standard-normal transformed features."),
    'solver': 'L-BFGS',
    'active': True,
    'keywords': [
        'robust regression', 'univariate regression', 'multiple regression',
        'theil-sen regression', 'sklearn.linear_model.TheilSenRegressor',
        'theil-sen estimator', 'sen slope estimator', 'slope selection',
        'single-median method', 'kendall robust line-fit method',
        'kendall–theil robust line'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}



# Interpret CV results - Pick the Best
#    n_best  depends on cv_settings['n_iter']
#    lowest=False  means that max(score) is the "best"
import pandas as pd


def select_the_best(res, n_best=5, lowest=False):
    # copy into dataframe
    res = pd.DataFrame(res.cv_results_)

    # pick the N best
    res = res.sort_values(by='mean_test_score', ascending=lowest)[:n_best]

    # compute CV-ratio depending on the score type
    if lowest:  # small mean(cv), small std(cv)
        res['cvratio'] = -(res['mean_test_score'] * res['std_test_score'])
    else:  # high mean(cv), small std(cv)
        res['cvratio'] = res['mean_test_score'] / res['std_test_score']

    # pick the best CV-Ratio of the N best ("best" must be the highest)
    res = res.sort_values(by='cvratio', ascending=False)

    # prep output args
    bestparams = res['params'][res.index[0]]  # best params
    summary = res[[
        'params', 'cvratio', 'rank_test_score',
        'mean_test_score', 'std_test_score']]
    return bestparams, summary



from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import numpy as np


class PcaEV(Pipeline):
    """Principal component analysis (PCA) with the number of components
    set to reach a required explained variance.

    Parameters
    ----------
    required_ev : float (Default 0.1)
        Required Explained Variance threshold.

    Example
    -------
        trans = PcaEV()
        trans.set_params(**{'required_ev': 0.2})
        trans.fit(X_train)
        X_new = trans.transform(X_train)

    Notes
    -----
    The class 'PcaEV' is a sklearn Pipeline and is equivalent to

        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.decomposition import PCA
        required_ev = 0.1
        trans = Pipeline(steps=[
            ('scl', StandardScaler()),
            ('pca', PCA(n_components = required_ev, svd_solver='full'))
        ])
        trans.fit(X_train)
        X_new = trans.transform(X_train).astype(np.float16)

    PcaEV
        - runs checks that "0.0 < PCA.n_components < 1.0",
        - transformed outputs are memory-friendly (np.float16),
        - create "feature_names_" output
    """
    def __init__(self, required_ev=0.1, prefix="pcaev"):
        self.prefix = prefix
        self.required_ev = required_ev
        super().__init__(steps=[
            ('scl', StandardScaler(with_mean=True, with_std=True, copy=True)),
            ('pca', PCA(
                n_components=self.required_ev,
                svd_solver='full', whiten=False, copy=True))
        ])

    def __str__(self):
        return 'PcaEV(required_ev={}, steps={}, prefix={})'.format(
            self.required_ev, self.steps, self.prefix)

    def __repr__(self):
        return self.__str__()

    def set_params(self, **kwargs):
        if 'required_ev' in kwargs:
            self.required_ev = kwargs['required_ev']
            self.steps[1][1].set_params(**{'n_components': self.required_ev})

    def transform(self, X, y=None):
        return super().transform(X).astype(np.float16)

    def fit(self, X, y=None):
        super().fit(X)
        self.feature_names_ = [
            self.prefix + "_" + str(i) for i in
            range(self.steps[1][1].n_components_)]
        return self


trans = PcaEV(required_ev=0.1)

meta = {
    'id': 'dim3',
    'name': 'PCA req EV',
    'description': (
        "Number of components is determined by "
        "a required Explained Variance threshold"),
    'keywords': [
        'dimensionality reduction', 'principal component anlysis',
        'StandardScaler', 'PCA', 'Explained Variance']
}



# Refit
from copy import deepcopy


def refit_model(basemodel, newparams, X, y=None):
    newmodel = deepcopy(basemodel)
    newmodel.set_params(**newparams)
    newmodel.fit(X, y)
    return newmodel



# For Univariate Regression
from sklearn.metrics import (
    make_scorer, r2_score,
    explained_variance_score, mean_absolute_error, mean_squared_error,
    mean_squared_log_error, median_absolute_error)

# Create Scorer - Ensure the lowest value means "best"
scorerfun = r2_score

# Hyperparameter Optimization settings
cv_settings = {
    'n_iter': 50,  # How many random hyperparmeter settings to try
    'cv': 5,  # Number of CV folds
    'scoring': make_scorer(scorerfun),
    'random_state': 42,
    'n_jobs': -1,
    'return_train_score': True
}


# Evaluate by watching some metrics
def print_scores(y_true, y_pred):
    """
    Y_pred = model.predict(X_valid)
    print_scores(Y_valid, Y_pred)
    """
    print("{:>20s}: {:12.4f}".format(
        "Mean Squared Error", mean_squared_error(y_true, y_pred)))
    print("{:>20s}: {:12.4f}".format(
        "Mean Absolute Error", mean_absolute_error(y_true, y_pred)))
    print("{:>20s}: {:12.4f}".format(
        "Median Abs Error", median_absolute_error(y_true, y_pred)))
    # print("{:>20s}: {:12.4f}".format(
    #    "Mean Sq Log Error", mean_squared_log_error(y_true, y_pred)))
    print("{:>20s}: {:12.4f}".format(
        "R Squared", r2_score(y_true, y_pred)))
    print("{:>20s}: {:12.4f}".format(
        "Explained Variance", explained_variance_score(y_true, y_pred)))




from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import ElasticNet
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', ElasticNet(
        fit_intercept = True,
        positive = False,
        random_state = 42,
        selection = 'random',
        precompute = True,
        tol = 0.0001,
        max_iter = 1000,
        warm_start = True,
        normalize = False,  # done in the pipeline
        copy_X = True,
    ))
])

hyper = {
    'lin__alpha': ss.gamma(a=2.2, loc=1e-3, scale=0.7),
    'lin__l1_ratio': ss.uniform(0, 1),
}

meta = {
    'id': "simi4",
    'name': 'LinReg ElasticNet',
    'descriptions': (
        "ElasticNet Regression (L1/L2 penalty), Coordinate Descent solver."),
    'solver': 'Coordinate Descent',
    'active': True,
    'keywords': [
        'linear regression', 'univariate regression', 'multiple regression',
        'elasticnet', 'sklearn.linear_model.ElasticNet'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}




from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import Ridge
import scipy.stats as ss

model = Pipeline(steps=[
    ('scl', StandardScaler()),
    ('lin', Ridge(
        solver = 'sag',
        tol = 0.001,  # optimizer termination criteria
        # alpha = 1.0,  # L2 regulization alpha=C^{-1}
        fit_intercept = True,
        normalize = False,  # done in the pipeline
        copy_X = True,
        random_state = 42,   # for SAG/SAGA solver
    ))
])

hyper = {
    'lin__alpha': ss.gamma(a=1.5, loc=1e-5, scale=.7),  # alpha ~ [0.001, 10]
}

meta = {
    'id': "simi7",
    'name': 'LinReg Ridge',
    'descriptions': (
        "Ridge Regression (L2 penalty), SAG solver, standard-normal "
        "transformed features."),
    'solver': 'Stochastic Average Gradient descent (SAG)',
    'active': True,
    'keywords': [
        'linear regression', 'univariate regression', 'multiple regression',
        'ridge', 'l2 penalty', 'sklearn.linear_model.Ridge'],
    'output_num': 'single',
    'output_scale': 'interval',
    'output_dtype': 'float',
    'input_num': 'multi',
    'input_scale': 'interval',
    'input_dtype': 'float'
}
